/*\

Copyright 2017 The University of North Carolina at Chapel Hill.
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation for educational, research and non-profit purposes, without
fee, and without a written agreement is hereby granted, provided that the
above copyright notice and the following three paragraphs appear in all
copies.

IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL BE
LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY
OF NORTH CAROLINA HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH
DAMAGES.

THE UNIVERSITY OF NORTH CAROLINA SPECIFICALLY DISCLAIM ANY
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
NORTH CAROLINA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

The authors may be contacted via:

US Mail: Real-Time Systems Group at UNC
Department of Computer Science
Sitterson Hall
University of N. Carolina
Chapel Hill, NC 27599-3175

EMail: nemitz@cs.unc.edu; tamert@cs.unc.edu; anderson@cs.unc.edu

**/

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include "rnlp.h"

/*
 *  RNLP: initialize the system
 */
void rnlp_init(rnlp_lock_t* lock)
{
    lock->head = NULL;
    lock->tail = NULL;
    lock->locked = 0;
    lock->unavailable = 0;
    ticket_init(&(lock->sublock));
}

/*
 *  RNLP: lock
 */
void rnlp_lock(rnlp_lock_t* lock, rnlp_node_t* self)
{
    ticket_lock(&(lock->sublock));

    if (0 == (lock->unavailable & self->requested))
    {
        self->blocked = 0;
        lock->locked |= self->requested;
        lock->unavailable |= self->requested;
    }
    else if (NULL == lock->head)
    {
        self->blocked = 1;
        lock->head = self;
        lock->tail = self;
        self->next = NULL;
    }
    else
    {
        self->blocked = 1;
        lock->tail->next = self;
        self->next = NULL;
        lock->tail = self;
    }

    ticket_unlock(&(lock->sublock));

    while (self->blocked == 1)
    {
        cpu_relax();
    }
}

/*
 *  RNLP: unlock
 */
void rnlp_unlock(rnlp_lock_t* lock, rnlp_node_t* self)
{
    rnlp_node_t *tmp, *prev = NULL;

    ticket_lock(&(lock->sublock));

    lock->locked &= ~(self->requested);
    lock->unavailable = lock->locked;

    tmp = (rnlp_node_t*)lock->head;

    while (tmp != NULL)
    {
        assert(tmp->blocked != 0);

        if (0 == (lock->unavailable & tmp->requested))
        {
            tmp->blocked = 0;
            lock->locked |= tmp->requested;

            lock->unavailable |= tmp->requested;

            if (tmp == lock->head)
            {
                lock->head = tmp->next;
                tmp = tmp->next;
            }
            else if (tmp == lock->tail)
            {
                prev->next = NULL;
                lock->tail = prev;
                tmp = NULL;
            }
            else
            {
                prev->next = tmp->next;
                tmp = tmp->next;
            }
        }
        else
        {
            lock->unavailable |= tmp->requested;
            prev = tmp;
            tmp = tmp->next;
        }
    }

    ticket_unlock(&(lock->sublock));
}

